from decimal import *
getcontext().prec = 8

import xml.etree.ElementTree as ET


def try_str(valor):
    """
    Intenta convertir un valor a texto,
    si es caracter alfanumérico lo convierte, sino devuelve vacío ''
    """
    if valor:
        return str(valor)
    else:
        return ''


def try_int(number):
    """
    Intenta convertir a entero una variable,
    si es un número lo convierte, sino devuelve None
    """
    try:
        return int(number)
    except:
        return None

def int_or_same(number):
    """
    Intenta convertir a entero una variable,
    si es un número lo convierte, sino devuelve el mismo valor
    Se hace esto porque Pandas agrega el .0 a los números
    """
    try:
        return int(number)
    except Exception:
        return number


def int_or_zero(number):
    """
    Intenta convertir a entero una variable,
    si es un número lo convierte, sino devuelve 0
    """
    try:
        return int(number)
    except Exception:
        return 0

def int_or_empty(number):
    """
    Intenta convertir a entero una variable,
    si es un número lo convierte, sino devuelve ''
    """
    try:
        return int(number)
    except Exception:
        return ''

def decimal_or_zero(number):
    """
    Intenta convertir a decimal una variable,
    si es un número lo convierte, sino devuelve 0
    """
    try:
        return round(float(number), 8)
    except Exception as error:
        return 0


def get_dict_etiq():
    """
    Mapeo de las todas las etiquetas mayúsculas con el original.
    Cuando se consulta a una vista Oracle devuelve todo en mayúscula
    Devuelve el diccionario
    """
    dic_etiquetas = {
        'DVERFOR': 'dVerFor',
        #'DE': 'DE',
        #'ID': 'id',
        'ID': 'id',
        'DE_ID': 'de_id',
        'DDVID': 'dDVId',
        'DFECFIRMA': 'dFecFirma',
        'DSISFACT': 'dSisFact',
        'GOPEDE': 'gOpeDE',
        'ITIPEMI': 'iTipEmi',
        'DDESTIPEMI': 'dDesTipEmi',
        'DCODSEG': 'dCodSeg',
        'DINFOEMI': 'dInfoEmi',
        'DINFOFISC': 'dInfoFisc',
        'GTIMB': 'gTimb',
        'ITIDE': 'iTiDE',
        'DDESTIDE': 'dDesTiDE',
        'DNUMTIM': 'dNumTim',
        'DEST': 'dEst',
        'DPUNEXP': 'dPunExp',
        'DNUMDOC': 'dNumDoc',
        'DSERIENUM': 'dSerieNum',
        'DFEINIT': 'dFeIniT',
        'DFEFINT': 'dFeFinT',
        'GDATGRALOPE': 'gDatGralOpe',
        'DFEEMIDE': 'dFeEmiDE',
        'GOPECOM': 'gOpeCom',
        'ITIPTRA': 'iTipTra',
        'DDESTIPTRA': 'dDesTipTra',
        'ITIMP': 'iTImp',
        'DDESTIMP': 'dDesTImp',
        'CMONEOPE': 'cMoneOpe',
        'DDESMONEOPE': 'dDesMoneOpe',
        'DCONDTICAM': 'dCondTiCam',
        'DTICAM': 'dTiCam',
        'ICONDANT': 'iCondAnt',
        'DDESCONDANT': 'dDesCondAnt',
        'GEMIS': 'gEmis',
        'DRUCEM': 'dRucEm',
        'DDVEMI': 'dDVEmi',
        'ITIPCONT': 'iTipCont',
        'CTIPREG': 'cTipReg',
        'DNOMEMI': 'dNomEmi',
        'DNOMFANEMI': 'dNomFanEmi',
        'DDIREMI': 'dDirEmi',
        'DNUMCAS': 'dNumCas',
        'DCOMPDIR1': 'dCompDir1',
        'DCOMPDIR2': 'dCompDir2',
        'CDEPEMI': 'cDepEmi',
        'DDESDEPEMI': 'dDesDepEmi',
        'CDISEMI': 'cDisEmi',
        'DDESDISEMI': 'dDesDisEmi',
        'CCIUEMI': 'cCiuEmi',
        'DDESCIUEMI': 'dDesCiuEmi',
        'DTELEMI': 'dTelEmi',
        'DEMAILE': 'dEmailE',
        'DDENSUC': 'dDenSuc',
        'GACTECO': 'gActEco',
        'CACTECO': 'cActEco',
        'DDESACTECO': 'dDesActEco',
        'GDATREC': 'gDatRec',
        'INATREC': 'iNatRec',
        'ITIOPE': 'iTiOpe',
        'CPAISREC': 'cPaisRec',
        'DDESPAISRE': 'dDesPaisRe',
        'ITICONTREC': 'iTiContRec',
        'DRUCREC': 'dRucRec',
        'DDVREC': 'dDVRec',
        'ITIPIDREC': 'iTipIDRec',
        'DDTIPIDREC': 'dDTipIDRec',
        'DNUMIDREC': 'dNumIDRec',
        'DNOMREC': 'dNomRec',
        'DNOMFANREC': 'dNomFanRec',
        'DDIRREC': 'dDirRec',
        'DNUMCASREC': 'dNumCasRec',
        'CDEPREC': 'cDepRec',
        'DDESDEPREC': 'dDesDepRec',
        'CDISREC': 'cDisRec',
        'DDESDISREC': 'dDesDisRec',
        'CCIUREC': 'cCiuRec',
        'DDESCIUREC': 'dDesCiuRec',
        'DTELREC': 'dTelRec',
        'DCELREC': 'dCelRec',
        'DEMAILREC': 'dEmailRec',
        'DCODCLIENTE': 'dCodCliente',
        'GDTIPDE': 'gDtipDE',
        'GCAMFE': 'gCamFE',
        'IINDPRES': 'iIndPres',
        'DDESINDPRES': 'dDesIndPres',
        'DFECEMNR': 'dFecEmNR',
        'GCOMPPUB': 'gCompPub',
        'DMODCONT': 'dModCont',
        'DENTCONT': 'dEntCont',
        'DANOCONT': 'dAnoCont',
        'DSECCONT': 'dSecCont',
        'DFECODCONT': 'dFeCodCont',
        'GCAMAE': 'gCamAE',
        'INATVEN': 'iNatVen',
        'DDESNATVEN': 'dDesNatVen',
        'ITIPIDVEN': 'iTipIDVen',
        'DDTIPIDVEN': 'dDTipIDVen',
        'DNUMIDVEN': 'dNumIDVen',
        'DNOMVEN': 'dNomVen',
        'DDIRVEN': 'dDirVen',
        'DNUMCASVEN': 'dNumCasVen',
        'CDEPVEN': 'cDepVen',
        'DDESDEPVEN': 'dDesDepVen',
        'CDISVEN': 'cDisVen',
        'DDESDISVEN': 'dDesDisVen',
        'CCIUVEN': 'cCiuVen',
        'DDESCIUVEN': 'dDesCiuVen',
        'DDIRPROV ': 'dDirProv ',
        'CDEPPROV ': 'cDepProv ',
        'DDESDEPPROV ': 'dDesDepProv ',
        'CDISPROV ': 'cDisProv ',
        'DDESDISPROV ': 'dDesDisProv ',
        'CCIUPROV ': 'cCiuProv ',
        'DDESCIUPROV ': 'dDesCiuProv ',
        'GCAMNCDE': 'gCamNCDE',
        'IMOTEMI': 'iMotEmi',
        'DDESMOTEMI': 'dDesMotEmi',
        'GCAMNRE': 'gCamNRE',
        'IMOTEMINR': 'iMotEmiNR',
        'DDESMOTEMINR': 'dDesMotEmiNR',
        'IRESPEMINR': 'iRespEmiNR',
        'DDESRESPEMINR': 'dDesRespEmiNR',
        'DKMR': 'dKmR',
        'DFECEM': 'dFecEm',
        'GCAMCOND': 'gCamCond',
        'ICONDOPE': 'iCondOpe',
        'DDCONDOPE': 'dDCondOpe',
        'GPACONEINI': 'gPaConEIni',
        'ITIPAGO': 'iTiPago',
        'DDESTIPAG': 'dDesTiPag',
        'DMONTIPAG': 'dMonTiPag',
        'CMONETIPAG': 'cMoneTiPag',
        'DDMONETIPAG': 'dDMoneTiPag',
        'DTICAMTIPAG': 'dTiCamTiPag',
        'GPAGTARCD': 'gPagTarCD',
        'IDENTARJ': 'iDenTarj',
        'DDESDENTARJ': 'dDesDenTarj',
        'DRSPROTAR': 'dRSProTar',
        'DRUCPROTAR': 'dRUCProTar',
        'DDVPROTAR': 'dDVProTar',
        'IFORPROPA': 'iForProPa',
        'DCODAUOPE': 'dCodAuOpe',
        'DNOMTIT': 'dNomTit',
        'DNUMTARJ': 'dNumTarj',
        'GPAGCHEQ': 'gPagCheq',
        'DNUMCHEQ': 'dNumCheq',
        'DBCOEMI': 'dBcoEmi',
        'GPAGCRED': 'gPagCred',
        'ICONDCRED': 'iCondCred',
        'DDCONDCRED': 'dDCondCred',
        'DPLAZOCRE': 'dPlazoCre',
        'DCUOTAS': 'dCuotas',
        'DMONENT': 'dMonEnt',
        'GCUOTAS': 'gCuotas',
        'CMONECUO': 'cMoneCuo',
        'DDMONECUO': 'dDMoneCuo',
        'DMONCUOTA': 'dMonCuota',
        'DVENCCUO': 'dVencCuo',
        'GCAMITEM': 'gCamItem',
        'DCODINT': 'dCodInt',
        'DPARARANC': 'dParAranc',
        'DNCM': 'dNCM',
        'DDNCPG': 'dDncpG',
        'DDNCPE': 'dDncpE',
        'DGTIN': 'dGtin',
        'DGTINPQ': 'dGtinPq',
        'DDESPROSER': 'dDesProSer',
        'CUNIMED': 'cUniMed',
        'DDESUNIMED': 'dDesUniMed',
        'DCANTPROSER': 'dCantProSer',
        'CPAISORIG': 'cPaisOrig',
        'DDESPAISORIG': 'dDesPaisOrig',
        'DINFITEM': 'dInfItem',
        'CRELMERC': 'cRelMerc',
        'DDESRELMERC': 'dDesRelMerc',
        'DCANQUIMER': 'dCanQuiMer',
        'DPORQUIMER': 'dPorQuiMer',
        'DCDCANTICIPO': 'dCDCAnticipo',
        'GVALORITEM': 'gValorItem',
        'DPUNIPROSER': 'dPUniProSer',
        'DTICAMIT': 'dTiCamIt',
        'DTOTBRUOPEITEM': 'dTotBruOpeItem',
        'GVALORRESTAITEM': 'gValorRestaItem',
        'DDESCITEM': 'dDescItem',
        'DPORCDESIT': 'dPorcDesIt',
        'DDESCGLOITEM': 'dDescGloItem',
        'DANTPREUNIIT': 'dAntPreUniIt',
        'DANTGLOPREUNIIT': 'dAntGloPreUniIt',
        'DTOTOPEITEM': 'dTotOpeItem',
        'DTOTOPEGS': 'dTotOpeGs',
        'GCAMIVA': 'gCamIVA',
        'IAFECIVA': 'iAfecIVA',
        'DDESAFECIVA': 'dDesAfecIVA',
        'DPROPIVA': 'dPropIVA',
        'DTASAIVA': 'dTasaIVA',
        'DBASGRAVIVA': 'dBasGravIVA',
        'DLIQIVAITEM': 'dLiqIVAItem',
        'GRASMERC': 'gRasMerc',
        'DNUMLOTE': 'dNumLote',
        'DVENCMERC': 'dVencMerc',
        'DNSERIE': 'dNSerie',
        'DNUMPEDI': 'dNumPedi',
        'DNUMSEGUI': 'dNumSegui',
        'DNOMIMP': 'dNomImp',
        'DDIRIMP': 'dDirImp',
        'DNUMFIR': 'dNumFir',
        'DNUMREG': 'dNumReg',
        'GCAMESP': 'gCamEsp',
        'GVEHNUEVO': 'gVehNuevo',
        'ITIPOPVN': 'iTipOpVN',
        'DDESTIPOPVN': 'dDesTipOpVN',
        'DCHASIS': 'dChasis',
        'DCOLOR': 'dColor',
        'DPOTENCIA': 'dPotencia',
        'DCAPMOT': 'dCapMot',
        'DPNET': 'dPNet',
        'DPBRUTO': 'dPBruto',
        'ITIPCOM': 'iTipCom',
        'DDESTIPCOM': 'dDesTipCom',
        'DNROMOTOR': 'dNroMotor',
        'DCAPTRACC': 'dCapTracc',
        'DANOFAB': 'dAnoFab',
        'CTIPVEH': 'cTipVeh',
        'DCAPAC': 'dCapac',
        'GGRUPENER': 'gGrupEner',
        'DNROMED': 'dNroMed',
        'DACTIV': 'dActiv',
        'DCATEG': 'dCateg',
        'DLECANT': 'dLecAnt',
        'DLECACT': 'dLecAct',
        'DCONKWH': 'dConKwh',
        'GGRUPSEG': 'gGrupSeg',
        'DCODEMPSEG': 'dCodEmpSeg',
        'GGRUPPOLSEG': 'gGrupPolSeg',
        'DPOLIZA': 'dPoliza',
        'DUNIDVIG': 'dUnidVig',
        'DVIGENCIA': 'dVigencia',
        'DNUMPOLIZA': 'dNumPoliza',
        'DCODINT': 'dCodInt',
        'GGRUPSUP': 'gGrupSup',
        'DNOMCAJ': 'dNomCaj',
        'DEFECTIVO': 'dEfectivo',
        'DVUELTO': 'dVuelto',
        'DDONAC': 'dDonac',
        'DDESDONAC': 'dDesDonac',
        'GGRUPADI': 'gGrupAdi',
        'DCICLO': 'dCiclo',
        'DFECINIC': 'dFecIniC',
        'DFECFINC': 'dFecFinC',
        'DVENCPAG': 'dVencPag',
        'DCONTRATO': 'dContrato',
        'DSALANT': 'dSalAnt',
        'GTRANSP': 'gTransp',
        'ITIPTRANS': 'iTipTrans',
        'DDESTIPTRANS': 'dDesTipTrans',
        'IMODTRANS': 'iModTrans',
        'DDESMODTRANS': 'dDesModTrans',
        'IRESPFLETE': 'iRespFlete',
        'CCONDNEG': 'cCondNeg',
        'DNUMANIF': 'dNuManif',
        'DNUDESPIMP': 'dNuDespImp',
        'DINITRAS': 'dIniTras',
        'DFINTRAS': 'dFinTras',
        'CPAISDEST': 'cPaisDest',
        'DDESPAISDEST': 'dDesPaisDest',
        'GCAMSAL': 'gCamSal',
        'DDIRLOCSAL': 'dDirLocSal',
        'DNUMCASSAL': 'dNumCasSal',
        'DCOMP1SAL': 'dComp1Sal',
        'DCOMP2SAL': 'dComp2Sal',
        'CDEPSAL': 'cDepSal',
        'DDESDEPSAL': 'dDesDepSal',
        'CDISSAL': 'cDisSal',
        'DDESDISSAL': 'dDesDisSal',
        'CCIUSAL': 'cCiuSal',
        'DDESCIUSAL': 'dDesCiuSal',
        'DTELSAL': 'dTelSal',
        'GCAMENT': 'gCamEnt',
        'DDIRLOCENT': 'dDirLocEnt',
        'DNUMCASENT': 'dNumCasEnt',
        'DCOMP1ENT': 'dComp1Ent',
        'DCOMP2ENT': 'dComp2Ent',
        'CDEPENT': 'cDepEnt',
        'DDESDEPENT': 'dDesDepEnt',
        'CDISENT': 'cDisEnt',
        'DDESDISENT': 'dDesDisEnt',
        'CCIUENT': 'cCiuEnt',
        'DDESCIUENT': 'dDesCiuEnt',
        'DTELENT': 'dTelEnt',
        'GVEHTRAS': 'gVehTras',
        'DTIVEHTRAS': 'dTiVehTras',
        'DMARVEH': 'dMarVeh',
        'DTIPIDENVEH': 'dTipIdenVeh',
        'DNROIDVEH': 'dNroIDVeh',
        'DADICVEH': 'dAdicVeh',
        'DNROMATVEH': 'dNroMatVeh',
        'DNROVUELO': 'dNroVuelo',
        'GCAMTRANS': 'gCamTrans',
        'INATTRANS': 'iNatTrans',
        'DNOMTRANS': 'dNomTrans',
        'DRUCTRANS': 'dRucTrans',
        'DDVTRANS': 'dDVTrans',
        'ITIPIDTRANS': 'iTipIDTrans',
        'DDTIPIDTRANS': 'dDTipIDTrans',
        'DNUMIDTRANS': 'dNumIDTrans',
        'CNACTRANS': 'cNacTrans',
        'DDESNACTRANS': 'dDesNacTrans',
        'DNUMIDCHOF': 'dNumIDChof',
        'DNOMCHOF': 'dNomChof',
        'DDOMFISC': 'dDomFisc',
        'DDIRCHOF': 'dDirChof',
        'DNOMBAG': 'dNombAg',
        'DRUCAG': 'dRucAg',
        'DDVAG': 'dDVAg',
        'DDIRAGE': 'dDirAge',
        'GTOTSUB': 'gTotSub',
        'DSUBEXE': 'dSubExe',
        'DSUBEXO': 'dSubExo',
        'DSUB5': 'dSub5',
        'DSUB10': 'dSub10',
        'DTOTOPE': 'dTotOpe',
        'DTOTDESC': 'dTotDesc',
        'DTOTDESCGLOTEM': 'dTotDescGlotem',
        'DTOTANTITEM': 'dTotAntItem',
        'DTOTANT': 'dTotAnt',
        'DPORCDESCTOTAL': 'dPorcDescTotal',
        'DDESCTOTAL': 'dDescTotal',
        'DANTICIPO': 'dAnticipo',
        'DREDON': 'dRedon',
        'DCOMI': 'dComi',
        'DTOTGRALOPE': 'dTotGralOpe',
        'DIVA5': 'dIVA5',
        'DIVA10': 'dIVA10',
        'DLIQTOTIVA5': 'dLiqTotIVA5',
        'DLIQTOTIVA10': 'dLiqTotIVA10',
        'DIVACOMI': 'dIVAComi',
        'DTOTIVA': 'dTotIVA',
        'DBASEGRAV5': 'dBaseGrav5',
        'DBASEGRAV10': 'dBaseGrav10',
        'DTBASGRAIVA': 'dTBasGraIVA',
        'DTOTALGS': 'dTotalGs',
        'DTOTCOM': 'dTotCom',
        'GCAMGEN': 'gCamGen',
        'DORDCOMPRA': 'dOrdCompra',
        'DORDVTA': 'dOrdVta',
        'DASIENTO': 'dAsiento',
        'GCAMCARG': 'gCamCarg',
        'CUNIMEDTOTVOL': 'cUniMedTotVol',
        'DDESUNIMEDTOTVOL': 'dDesUniMedTotVol',
        'DTOTVOLMERC': 'dTotVolMerc',
        'CUNIMEDTOTPES': 'cUniMedTotPes',
        'DDESUNIMEDTOTPES': 'dDesUniMedTotPes',
        'DTOTPESMERC': 'dTotPesMerc',
        'ICARCARGA': 'iCarCarga',
        'DDESCARCARGA': 'dDesCarCarga',
        'GCAMDEASOC': 'gCamDEAsoc',
        'ITIPDOCASO': 'iTipDocAso',
        'DDESTIPDOCASO': 'dDesTipDocAso',
        'DCDCDEREF': 'dCdCDERef',
        'DNTIMDI': 'dNTimDI',
        'DESTDOCASO': 'dEstDocAso',
        'DPEXPDOCASO': 'dPExpDocAso',
        'DNUMDOCASO': 'dNumDocAso',
        'ITIPODOCASO': 'iTipoDocAso',
        'DDTIPODOCASO': 'dDTipoDocAso',
        'DFECEMIDI': 'dFecEmiDI',
        'DNUMCOMRET': 'dNumComRet',
        'DNUMRESCF': 'dNumResCF',
        'ITIPCONS': 'iTipCons',
        'DDESTIPCONS': 'dDesTipCons',
        'DNUMCONS': 'dNumCons',
        'DNUMCONTROL': 'dNumControl',
        'SIGNATURE': 'Signature',
        'GCAMFUFD': 'gCamFuFD',
        'DCARQR': 'dCarQR',
        'DINFADIC': 'dInfAdic',

        'dverfor': 'dVerFor',
         #'de': 'DE',
         #'id': 'id',
        'ddvid': 'dDVId',
        'dfecfirma': 'dFecFirma',
        'dsisfact': 'dSisFact',
        'gopede': 'gOpeDE',
        'itipemi': 'iTipEmi',
        'ddestipemi': 'dDesTipEmi',
        'dcodseg': 'dCodSeg',
        'dinfoemi': 'dInfoEmi',
        'dinfofisc': 'dInfoFisc',
        'gtimb': 'gTimb',
        'itide': 'iTiDE',
        'ddestide': 'dDesTiDE',
        'dnumtim': 'dNumTim',
        'dest': 'dEst',
        'dpunexp': 'dPunExp',
        'dnumdoc': 'dNumDoc',
        'dserienum': 'dSerieNum',
        'dfeinit': 'dFeIniT',
        'dfefint': 'dFeFinT',
        'gdatgralope': 'gDatGralOpe',
        'dfeemide': 'dFeEmiDE',
        'gopecom': 'gOpeCom',
        'itiptra': 'iTipTra',
        'ddestiptra': 'dDesTipTra',
        'itimp': 'iTImp',
        'ddestimp': 'dDesTImp',
        'cmoneope': 'cMoneOpe',
        'ddesmoneope': 'dDesMoneOpe',
        'dcondticam': 'dCondTiCam',
        'dticam': 'dTiCam',
        'icondant': 'iCondAnt',
        'ddescondant': 'dDesCondAnt',
        'gemis': 'gEmis',
        'drucem': 'dRucEm',
        'ddvemi': 'dDVEmi',
        'itipcont': 'iTipCont',
        'ctipreg': 'cTipReg',
        'dnomemi': 'dNomEmi',
        'dnomfanemi': 'dNomFanEmi',
        'ddiremi': 'dDirEmi',
        'dnumcas': 'dNumCas',
        'dcompdir1': 'dCompDir1',
        'dcompdir2': 'dCompDir2',
        'cdepemi': 'cDepEmi',
        'ddesdepemi': 'dDesDepEmi',
        'cdisemi': 'cDisEmi',
        'ddesdisemi': 'dDesDisEmi',
        'cciuemi': 'cCiuEmi',
        'ddesciuemi': 'dDesCiuEmi',
        'dtelemi': 'dTelEmi',
        'demaile': 'dEmailE',
        'ddensuc': 'dDenSuc',
        'gacteco': 'gActEco',
        'cacteco': 'cActEco',
        'ddesacteco': 'dDesActEco',
        'gdatrec': 'gDatRec',
        'inatrec': 'iNatRec',
        'itiope': 'iTiOpe',
        'cpaisrec': 'cPaisRec',
        'ddespaisre': 'dDesPaisRe',
        'iticontrec': 'iTiContRec',
        'drucrec': 'dRucRec',
        'ddvrec': 'dDVRec',
        'itipidrec': 'iTipIDRec',
        'ddtipidrec': 'dDTipIDRec',
        'dnumidrec': 'dNumIDRec',
        'dnomrec': 'dNomRec',
        'dnomfanrec': 'dNomFanRec',
        'ddirrec': 'dDirRec',
        'dnumcasrec': 'dNumCasRec',
        'cdeprec': 'cDepRec',
        'ddesdeprec': 'dDesDepRec',
        'cdisrec': 'cDisRec',
        'ddesdisrec': 'dDesDisRec',
        'cciurec': 'cCiuRec',
        'ddesciurec': 'dDesCiuRec',
        'dtelrec': 'dTelRec',
        'dcelrec': 'dCelRec',
        'demailrec': 'dEmailRec',
        'dcodcliente': 'dCodCliente',
        'gdtipde': 'gDtipDE',
        'gcamfe': 'gCamFE',
        'iindpres': 'iIndPres',
        'ddesindpres': 'dDesIndPres',
        'dfecemnr': 'dFecEmNR',
        'gcomppub': 'gCompPub',
        'dmodcont': 'dModCont',
        'dentcont': 'dEntCont',
        'danocont': 'dAnoCont',
        'dseccont': 'dSecCont',
        'dfecodcont': 'dFeCodCont',
        'gcamae': 'gCamAE',
        'inatven': 'iNatVen',
        'ddesnatven': 'dDesNatVen',
        'itipidven': 'iTipIDVen',
        'ddtipidven': 'dDTipIDVen',
        'dnumidven': 'dNumIDVen',
        'dnomven': 'dNomVen',
        'ddirven': 'dDirVen',
        'dnumcasven': 'dNumCasVen',
        'cdepven': 'cDepVen',
        'ddesdepven': 'dDesDepVen',
        'cdisven': 'cDisVen',
        'ddesdisven': 'dDesDisVen',
        'cciuven': 'cCiuVen',
        'ddesciuven': 'dDesCiuVen',
        'ddirprov ': 'dDirProv ',
        'cdepprov ': 'cDepProv ',
        'ddesdepprov ': 'dDesDepProv ',
        'cdisprov ': 'cDisProv ',
        'ddesdisprov ': 'dDesDisProv ',
        'cciuprov ': 'cCiuProv ',
        'ddesciuprov ': 'dDesCiuProv ',
        'gcamncde': 'gCamNCDE',
        'imotemi': 'iMotEmi',
        'ddesmotemi': 'dDesMotEmi',
        'gcamnre': 'gCamNRE',
        'imoteminr': 'iMotEmiNR',
        'ddesmoteminr': 'dDesMotEmiNR',
        'irespeminr': 'iRespEmiNR',
        'ddesrespeminr': 'dDesRespEmiNR',
        'dkmr': 'dKmR',
        'dfecem': 'dFecEm',
        'gcamcond': 'gCamCond',
        'icondope': 'iCondOpe',
        'ddcondope': 'dDCondOpe',
        'gpaconeini': 'gPaConEIni',
        'itipago': 'iTiPago',
        'ddestipag': 'dDesTiPag',
        'dmontipag': 'dMonTiPag',
        'cmonetipag': 'cMoneTiPag',
        'ddmonetipag': 'dDMoneTiPag',
        'dticamtipag': 'dTiCamTiPag',
        'gpagtarcd': 'gPagTarCD',
        'identarj': 'iDenTarj',
        'ddesdentarj': 'dDesDenTarj',
        'drsprotar': 'dRSProTar',
        'drucprotar': 'dRUCProTar',
        'ddvprotar': 'dDVProTar',
        'iforpropa': 'iForProPa',
        'dcodauope': 'dCodAuOpe',
        'dnomtit': 'dNomTit',
        'dnumtarj': 'dNumTarj',
        'gpagcheq': 'gPagCheq',
        'dnumcheq': 'dNumCheq',
        'dbcoemi': 'dBcoEmi',
        'gpagcred': 'gPagCred',
        'icondcred': 'iCondCred',
        'ddcondcred': 'dDCondCred',
        'dplazocre': 'dPlazoCre',
        'dcuotas': 'dCuotas',
        'dmonent': 'dMonEnt',
        'gcuotas': 'gCuotas',
        'cmonecuo': 'cMoneCuo',
        'ddmonecuo': 'dDMoneCuo',
        'dmoncuota': 'dMonCuota',
        'dvenccuo': 'dVencCuo',
        'gcamitem': 'gCamItem',
        'dcodint': 'dCodInt',
        'dpararanc': 'dParAranc',
        'dncm': 'dNCM',
        'ddncpg': 'dDncpG',
        'ddncpe': 'dDncpE',
        'dgtin': 'dGtin',
        'dgtinpq': 'dGtinPq',
        'ddesproser': 'dDesProSer',
        'cunimed': 'cUniMed',
        'ddesunimed': 'dDesUniMed',
        'dcantproser': 'dCantProSer',
        'cpaisorig': 'cPaisOrig',
        'ddespaisorig': 'dDesPaisOrig',
        'dinfitem': 'dInfItem',
        'crelmerc': 'cRelMerc',
        'ddesrelmerc': 'dDesRelMerc',
        'dcanquimer': 'dCanQuiMer',
        'dporquimer': 'dPorQuiMer',
        'dcdcanticipo': 'dCDCAnticipo',
        'gvaloritem': 'gValorItem',
        'dpuniproser': 'dPUniProSer',
        'dticamit': 'dTiCamIt',
        'dtotbruopeitem': 'dTotBruOpeItem',
        'gvalorrestaitem': 'gValorRestaItem',
        'ddescitem': 'dDescItem',
        'dporcdesit': 'dPorcDesIt',
        'ddescgloitem': 'dDescGloItem',
        'dantpreuniit': 'dAntPreUniIt',
        'dantglopreuniit': 'dAntGloPreUniIt',
        'dtotopeitem': 'dTotOpeItem',
        'dtotopegs': 'dTotOpeGs',
        'gcamiva': 'gCamIVA',
        'iafeciva': 'iAfecIVA',
        'ddesafeciva': 'dDesAfecIVA',
        'dpropiva': 'dPropIVA',
        'dtasaiva': 'dTasaIVA',
        'dbasgraviva': 'dBasGravIVA',
        'dliqivaitem': 'dLiqIVAItem',
        'grasmerc': 'gRasMerc',
        'dnumlote': 'dNumLote',
        'dvencmerc': 'dVencMerc',
        'dnserie': 'dNSerie',
        'dnumpedi': 'dNumPedi',
        'dnumsegui': 'dNumSegui',
        'dnomimp': 'dNomImp',
        'ddirimp': 'dDirImp',
        'dnumfir': 'dNumFir',
        'dnumreg': 'dNumReg',
        'gcamesp': 'gCamEsp',
        'gvehnuevo': 'gVehNuevo',
        'itipopvn': 'iTipOpVN',
        'ddestipopvn': 'dDesTipOpVN',
        'dchasis': 'dChasis',
        'dcolor': 'dColor',
        'dpotencia': 'dPotencia',
        'dcapmot': 'dCapMot',
        'dpnet': 'dPNet',
        'dpbruto': 'dPBruto',
        'itipcom': 'iTipCom',
        'ddestipcom': 'dDesTipCom',
        'dnromotor': 'dNroMotor',
        'dcaptracc': 'dCapTracc',
        'danofab': 'dAnoFab',
        'ctipveh': 'cTipVeh',
        'dcapac': 'dCapac',
        'ggrupener': 'gGrupEner',
        'dnromed': 'dNroMed',
        'dactiv': 'dActiv',
        'dcateg': 'dCateg',
        'dlecant': 'dLecAnt',
        'dlecact': 'dLecAct',
        'dconkwh': 'dConKwh',
        'ggrupseg': 'gGrupSeg',
        'dcodempseg': 'dCodEmpSeg',
        'ggruppolseg': 'gGrupPolSeg',
        'dpoliza': 'dPoliza',
        'dunidvig': 'dUnidVig',
        'dvigencia': 'dVigencia',
        'dnumpoliza': 'dNumPoliza',
        'dcodint': 'dCodInt',
        'ggrupsup': 'gGrupSup',
        'dnomcaj': 'dNomCaj',
        'defectivo': 'dEfectivo',
        'dvuelto': 'dVuelto',
        'ddonac': 'dDonac',
        'ddesdonac': 'dDesDonac',
        'ggrupadi': 'gGrupAdi',
        'dciclo': 'dCiclo',
        'dfecinic': 'dFecIniC',
        'dfecfinc': 'dFecFinC',
        'dvencpag': 'dVencPag',
        'dcontrato': 'dContrato',
        'dsalant': 'dSalAnt',
        'gtransp': 'gTransp',
        'itiptrans': 'iTipTrans',
        'ddestiptrans': 'dDesTipTrans',
        'imodtrans': 'iModTrans',
        'ddesmodtrans': 'dDesModTrans',
        'irespflete': 'iRespFlete',
        'ccondneg': 'cCondNeg',
        'dnumanif': 'dNuManif',
        'dnudespimp': 'dNuDespImp',
        'dinitras': 'dIniTras',
        'dfintras': 'dFinTras',
        'cpaisdest': 'cPaisDest',
        'ddespaisdest': 'dDesPaisDest',
        'gcamsal': 'gCamSal',
        'ddirlocsal': 'dDirLocSal',
        'dnumcassal': 'dNumCasSal',
        'dcomp1sal': 'dComp1Sal',
        'dcomp2sal': 'dComp2Sal',
        'cdepsal': 'cDepSal',
        'ddesdepsal': 'dDesDepSal',
        'cdissal': 'cDisSal',
        'ddesdissal': 'dDesDisSal',
        'cciusal': 'cCiuSal',
        'ddesciusal': 'dDesCiuSal',
        'dtelsal': 'dTelSal',
        'gcament': 'gCamEnt',
        'ddirlocent': 'dDirLocEnt',
        'dnumcasent': 'dNumCasEnt',
        'dcomp1ent': 'dComp1Ent',
        'dcomp2ent': 'dComp2Ent',
        'cdepent': 'cDepEnt',
        'ddesdepent': 'dDesDepEnt',
        'cdisent': 'cDisEnt',
        'ddesdisent': 'dDesDisEnt',
        'cciuent': 'cCiuEnt',
        'ddesciuent': 'dDesCiuEnt',
        'dtelent': 'dTelEnt',
        'gvehtras': 'gVehTras',
        'dtivehtras': 'dTiVehTras',
        'dmarveh': 'dMarVeh',
        'dtipidenveh': 'dTipIdenVeh',
        'dnroidveh': 'dNroIDVeh',
        'dadicveh': 'dAdicVeh',
        'dnromatveh': 'dNroMatVeh',
        'dnrovuelo': 'dNroVuelo',
        'gcamtrans': 'gCamTrans',
        'inattrans': 'iNatTrans',
        'dnomtrans': 'dNomTrans',
        'dructrans': 'dRucTrans',
        'ddvtrans': 'dDVTrans',
        'itipidtrans': 'iTipIDTrans',
        'ddtipidtrans': 'dDTipIDTrans',
        'dnumidtrans': 'dNumIDTrans',
        'cnactrans': 'cNacTrans',
        'ddesnactrans': 'dDesNacTrans',
        'dnumidchof': 'dNumIDChof',
        'dnomchof': 'dNomChof',
        'ddomfisc': 'dDomFisc',
        'ddirchof': 'dDirChof',
        'dnombag': 'dNombAg',
        'drucag': 'dRucAg',
        'ddvag': 'dDVAg',
        'ddirage': 'dDirAge',
        'gtotsub': 'gTotSub',
        'dsubexe': 'dSubExe',
        'dsubexo': 'dSubExo',
        'dsub5': 'dSub5',
        'dsub10': 'dSub10',
        'dtotope': 'dTotOpe',
        'dtotdesc': 'dTotDesc',
        'dtotdescglotem': 'dTotDescGlotem',
        'dtotantitem': 'dTotAntItem',
        'dtotant': 'dTotAnt',
        'dporcdesctotal': 'dPorcDescTotal',
        'ddesctotal': 'dDescTotal',
        'danticipo': 'dAnticipo',
        'dredon': 'dRedon',
        'dcomi': 'dComi',
        'dtotgralope': 'dTotGralOpe',
        'diva5': 'dIVA5',
        'diva10': 'dIVA10',
        'dliqtotiva5': 'dLiqTotIVA5',
        'dliqtotiva10': 'dLiqTotIVA10',
        'divacomi': 'dIVAComi',
        'dtotiva': 'dTotIVA',
        'dbasegrav5': 'dBaseGrav5',
        'dbasegrav10': 'dBaseGrav10',
        'dtbasgraiva': 'dTBasGraIVA',
        'dtotalgs': 'dTotalGs',
        'dtotcom': 'dTotCom',
        'gcamgen': 'gCamGen',
        'dordcompra': 'dOrdCompra',
        'dordvta': 'dOrdVta',
        'dasiento': 'dAsiento',
        'gcamcarg': 'gCamCarg',
        'cunimedtotvol': 'cUniMedTotVol',
        'ddesunimedtotvol': 'dDesUniMedTotVol',
        'dtotvolmerc': 'dTotVolMerc',
        'cunimedtotpes': 'cUniMedTotPes',
        'ddesunimedtotpes': 'dDesUniMedTotPes',
        'dtotpesmerc': 'dTotPesMerc',
        'icarcarga': 'iCarCarga',
        'ddescarcarga': 'dDesCarCarga',
        'gcamdeasoc': 'gCamDEAsoc',
        'itipdocaso': 'iTipDocAso',
        'ddestipdocaso': 'dDesTipDocAso',
        'dcdcderef': 'dCdCDERef',
        'dntimdi': 'dNTimDI',
        'destdocaso': 'dEstDocAso',
        'dpexpdocaso': 'dPExpDocAso',
        'dnumdocaso': 'dNumDocAso',
        'itipodocaso': 'iTipoDocAso',
        'ddtipodocaso': 'dDTipoDocAso',
        'dfecemidi': 'dFecEmiDI',
        'dnumcomret': 'dNumComRet',
        'dnumrescf': 'dNumResCF',
        'itipcons': 'iTipCons',
        'ddestipcons': 'dDesTipCons',
        'dnumcons': 'dNumCons',
        'dnumcontrol': 'dNumControl',
        'signature': 'Signature',
        'gcamfufd': 'gCamFuFD',
        'dcarqr': 'dCarQR',
        'dinfadic': 'dInfAdic',

        'TIPO_DE': 'tipo_de',
        'SECCION': 'seccion',
        'ETIQUETA': 'etiqueta',
        'VALOR': 'valor',
        }

    return dic_etiquetas


def set_valor_etiqueta(obj, etiqueta, valor):
    # Setea el valor en los objetos si la etiqueta corresponde al objeto
    if (etiqueta in dir(obj) and 
            (valor or str(valor) == '0' or str(valor) == '0.0') and
            etiqueta != 'id'):
        if isinstance(valor, Decimal) and not valor:
            valor = valor.normalize()
        if isinstance(valor, float):
            if valor.is_integer():
                valor = int(valor)
        # Debug para saber el tipo de dato de una etiqueta
        """
        if etiqueta == 'dInfAdic':
            tipo = type(valor)
            valor = valor
            print(f'tipo: {type(valor)} valor: {valor}')
        """
        setattr(obj, etiqueta, valor)
    return obj


def set_constantes(df_constantes, obj):
    # Recorre las constantes y setea los valores 
    for row in df_constantes.to_dict(orient="records"):
        etiqueta = row['etiqueta']
        valor = row['valor']
        # Seteamos los valores en los objetos si la etiqueta corresponde al objeto
        obj = set_valor_etiqueta(obj, etiqueta, valor)

    return obj


def set_valores_de(obj, df_constantes, row):
    # Obtenemos los valores de las constantes
    for const_row in df_constantes.to_dict(orient="records"):
        etiqueta = const_row['etiqueta']
        valor = const_row['valor']
        # Seteamos los valores en los objetos si la etiqueta 
        # corresponde al objeto
        obj = set_valor_etiqueta(obj, etiqueta, valor)

    # Recorremos cada columna del registro para setear los valores
    for etiqueta, valor in row.items():
        obj = set_valor_etiqueta(obj, etiqueta, valor)

    return obj


def set_valores(obj, df_constantes, df_objects, is_list=True):
    list_objects = []
    # Seteamos los valores de los items
    for row in df_objects.to_dict(orient="records"):
        obj = type(obj)()
        # Seteamos constantes
        obj = set_constantes(df_constantes, obj)
        # Recorremos cada columna del registro para setear los valores
        for etiqueta, valor in row.items():
            obj = set_valor_etiqueta(obj, etiqueta, valor)
        if is_list:
            list_objects.append(obj)
    if is_list:
        return list_objects
    else:
        return obj
    
def isXmlValid(value):
    """
    Verifica si un valor no va a romper el xml
    Por ejemplo 
    BRAVO LONGARTGABRIEL JOS              
    es un nombre de receptor con caracter especial que no es válido
    """
    try:
        xml = f"<value>{value}</value>"
        ET.fromstring(xml)
    except ET.ParseError:
        return False
    return True
        